@@ -49,13 +49,15 @@ def make_pipeline(*steps, memory=None, verbose=False):
 
 
 def recursive_build_pipeline_from_cfg(  cfg, 
-                                        node_name_set, 
+                                        node_name_set,
                                         node_pairs, 
+                                        node_name,
                                         latest_node, 
                                         index, 
                                         parallel_index, 
                                         nodename2transformer, 
-                                        print_task_result):
+                                        pipeline_settings,
+                                        kafka_media):
     """
     Examples
     -----------
@@ -70,6 +72,7 @@ def recursive_build_pipeline_from_cfg(  cfg,
     
     steps = []
 
+
     for trans_cfg in cfg["pipeline"]:
         trans_name = trans_cfg.get("type")  if "type" in trans_cfg else ""
         trans_name = "pipeline" if "pipeline" in trans_cfg else trans_name
@@ -92,14 +95,16 @@ def recursive_build_pipeline_from_cfg(  cfg,
             if isinstance(latest_node,str):
                 #convert the _previous_node to a list
                 latest_nodes= [latest_node for i in range(len(trans_cfg["transformers"]))]
+                pre_node_names = [node_name for i in range(len(trans_cfg["transformers"]))]
             elif isinstance(latest_node, list):
                 latest_nodes = latest_node
+                pre_node_names = node_name
             else:
                 assert False, "latest_node is not str or list"
 
             parallel_dict_index = [i for i in range(len(latest_nodes))]
             # print(parallel_dict_index, latest_nodes, trans_cfg)
-            for i, latest_node, sub_trans_cfg in zip(parallel_dict_index, latest_nodes, trans_cfg["transformers"]):
+            for i, node_name, latest_node, sub_trans_cfg in zip(parallel_dict_index, pre_node_names, latest_nodes, trans_cfg["transformers"]):
  
                 if "type" in sub_trans_cfg:
                     sub_trans_name = sub_trans_cfg.get("type")
@@ -115,12 +120,15 @@ def recursive_build_pipeline_from_cfg(  cfg,
                     TRANSFORMERS = get_register(module_type)
                     # print(TRANSFORMERS)
                     transformer = build_transformer_from_cfg(sub_trans_cfg, TRANSFORMERS)
+                    transformer.pre_node_name = node_name
                     transformer.index = index
                     transformer.node_name = f"{str(index)}.{sub_trans_name}"
                     # transformer.__name__
                     transformer.description = description
                     transformer.breakpoint = sub_breakpoint
-                    transformer.print_task_result = print_task_result
+                    transformer.print_task_result = pipeline_settings.print_task_result
+                    transformer.exchange_media = pipeline_settings.exchange_media
+                    transformer.kafka_media = kafka_media
                     transformers_list.append(transformer)
                     # update nodename2transformer
                     nodename2transformer[transformer.node_name] = transformer
@@ -132,14 +140,16 @@ def recursive_build_pipeline_from_cfg(  cfg,
 
 
                 elif "pipeline" in sub_trans_cfg:
-                    transformer, node_name_set, _node_pairs, _latest_node, index, parallel_index, nodename2transformer = recursive_build_pipeline_from_cfg( sub_trans_cfg, 
+                    transformer, node_name_set, _node_pairs, node_name, _latest_node, index, parallel_index, nodename2transformer = recursive_build_pipeline_from_cfg( sub_trans_cfg, 
                                                                                                                                                             node_name_set, 
                                                                                                                                                             [], 
+                                                                                                                                                            node_name,
                                                                                                                                                             latest_node, 
                                                                                                                                                             index, 
                                                                                                                                                             parallel_index, 
                                                                                                                                                             nodename2transformer,
-                                                                                                                                                            print_task_result)
+                                                                                                                                                            pipeline_settings,
+                                                                                                                                                            kafka_media)
                     transformers_list.append(transformer)
                     parallel_node_pairs[f"process_no.{str(i)}"]=_node_pairs
                     parallel_latest_node.append(_latest_node)
@@ -157,7 +167,7 @@ def recursive_build_pipeline_from_cfg(  cfg,
                                                               FUNCTIONAL_TRANSFORMERS, 
                                                               transformers_list)
             parallel_transformer.breakpoint = breakpoint
-            parallel_transformer.print_task_result = print_task_result
+            parallel_transformer.print_task_result = pipeline_settings.print_task_result
             parallel_transformer.node_name = parallel_node_name
             steps.append(parallel_transformer)
             # update nodename2transformer
@@ -167,20 +177,23 @@ def recursive_build_pipeline_from_cfg(  cfg,
             
         elif trans_name=="pipeline":
             # recursive function help update node_pairs, latest_node and index
-            _pipeline, node_name_set, _node_pairs, latest_node, index, parallel_index, nodename2transformer= recursive_build_pipeline_from_cfg( trans_cfg, 
+            _pipeline, node_name_set, _node_pairs, node_name, latest_node, index, parallel_index, nodename2transformer= recursive_build_pipeline_from_cfg( trans_cfg, 
                                                                                                                                                 node_name_set, 
                                                                                                                                                 [], 
-                                                                                                                                                latest_node, 
+                                                                                                                                                node_name,
+                                                                                                                                                latest_node,
                                                                                                                                                 index, 
                                                                                                                                                 parallel_index, 
                                                                                                                                                 nodename2transformer,
-                                                                                                                                                print_task_result)
+                                                                                                                                                pipeline_settings,
+                                                                                                                                                kafka_media)
             # update steps
             node_pairs.append(_node_pairs)
             steps.append(_pipeline)
         
         else:
             index = index + 1
+            pre_node_name = node_name
             node_name = f"{str(index)}.{trans_name}"
             # breakpoint=trans_cfg.pop("breakpoint") if "breakpoint" in trans_cfg else 0
             description = trans_cfg.pop("description") if "description" in trans_cfg else ""
@@ -191,12 +204,16 @@ def recursive_build_pipeline_from_cfg(  cfg,
             transformer.node_name = node_name
             transformer.description = description
             transformer.breakpoint = breakpoint
-            transformer.print_task_result = print_task_result
+            transformer.print_task_result = pipeline_settings.print_task_result
+            transformer.exchange_media = pipeline_settings.exchange_media
+            transformer.kafka_media = kafka_media
             
             # update steps list
             steps.append(transformer)
             # update nodename2transformer
             nodename2transformer[node_name] = transformer
+            # set pre_node_name
+            transformer.pre_node_name = pre_node_name
             # update node_pairs
             if isinstance(latest_node, str):
                 node_pairs.append((latest_node, node_name))
@@ -212,7 +229,7 @@ def recursive_build_pipeline_from_cfg(  cfg,
     node_name_set = set([node_name for node_name in node_name_set if node_name!=""])
 
 
-    return make_pipeline(*steps), node_name_set, node_pairs, latest_node, index, parallel_index, nodename2transformer
+    return make_pipeline(*steps), node_name_set, node_pairs, node_name, latest_node, index, parallel_index, nodename2transformer
             
 
 def create_tranformers_dict(transformers):
@@ -240,6 +257,7 @@ class Pipeline(sklearn_pipeline):
         self.verbose = verbose
         self.pipe_start = pipe_start
         self.pipe_stop = pipe_stop
+        self.delete_kafka_topic = 1
         
     
     def _islice(self, target_list, start, end):
@@ -276,6 +294,9 @@ class Pipeline(sklearn_pipeline):
         new = copy.deepcopy(self)
         
         new.steps = self._islice(self.steps, self.pipe_start, self.pipe_stop)
+
+        new.steps[0][1].is_first_trans = 1
+        new.steps[-1][1].is_last_trans = 1
         
         return new
 
@@ -287,21 +308,40 @@ class Pipeline(sklearn_pipeline):
 
     def transform(self, X):
         Xt = X
-        for _, _, transformer in self._iter():
+
+        # to be coding
+        # creating kafka topic with tempname
+
+        for idx, _, transformer in self._iter():
+            
+            # print(type(transformer), "#####################")
             Xt = transformer.transform(Xt)
+            
+
             if hasattr(transformer, "breakpoint"):
                 if transformer.breakpoint:
                     # print(transform.node_name, "has breakpoint", transform.breakpoint)
                     datetime_str = get_datetime_str(format="%Y-%m-%d %H:%M:%S")
                     print(f"{datetime_str} stop with breakpoint at {transformer.node_name}") if transformer.print_task_result else ""
                     return Xt
+        
+        # to be coding
+        if self.delete_kafka_topic:
+            # delete kafka topic
+            pass 
+
         return Xt
     
 
 
 class pipeline_settings():
     def __init__(self, cfg):
+
         self.pipeline_settings = cfg.pop("pipeline_settings") if "pipeline_settings" in cfg else {"print_cfg":0, "print_task_result":0}
+        self.print_cfg, self.print_task_result = self.get_pipeline_settings_from_cfg()
+        self.exchange_media = self.pipeline_settings["exchange_media"] if "exchange_media" in self.pipeline_settings else "default"
+        self.delete_kafka_topic = self.pipeline_settings["delete_kafka_topic"] if "delete_kafka_topic" in self.pipeline_settings else 1
+        self.bootstrap_servers = self.pipeline_settings["bootstrap_servers"] if "bootstrap_servers" in self.pipeline_settings else []
     
     def get_pipeline_settings_from_cfg(self):
         
@@ -324,15 +364,22 @@ class DataPipeline:
     """
     def __init__(self, cfg):
         #------------------------- pipeline_settings 
-        _pipeline_settings = pipeline_settings(cfg)
+        self._pipeline_settings = pipeline_settings(cfg)
         self._print_cfg, \
-        self._print_task_result = _pipeline_settings.get_pipeline_settings_from_cfg()
+        self._print_task_result = self._pipeline_settings.get_pipeline_settings_from_cfg()
 
         if self._print_cfg:
             print("".join(["-" for i in range(80)]), "pipeline configuration", "".join(["-" for i in range(80)]))
             print(cfg)
             print("".join(["-" for i in range(80)]), "----------------------", "".join(["-" for i in range(80)]))
 
+        # self.kafka_media = kafka_media
+        from .kafka_media import kafkaMedia
+        self.kafka_media = kafkaMedia(bootstrap_servers=self._pipeline_settings.bootstrap_servers, topic="pipeline")
+        self.kafka_media.__enter__()
+        # print(self.kafka_media.producer, "#############################")
+
+
         self._cfg = cfg
         self._node_name_set=set()
         self._node_pairs = []
@@ -376,17 +423,26 @@ class DataPipeline:
         _pipeline, \
         self._node_name_set, \
         self._node_pairs, \
+        _note_name, \
         self._latest_node, \
         index, \
         parallel_index, \
         self._nodename2transformer= recursive_build_pipeline_from_cfg(  self._cfg, 
                                                                         self._node_name_set,
                                                                         self._node_pairs, 
+                                                                        "",
                                                                         self._latest_node, 
                                                                         self._index,
                                                                         self._parallel_index,
                                                                         self._nodename2transformer,
-                                                                        self._print_task_result)
+                                                                        self._pipeline_settings,
+                                                                        self.kafka_media)
+
+        # setting first and last trans
+        if _pipeline.steps:
+            _pipeline.steps[0][1].is_first_trans = 1
+            _pipeline.steps[-1][1].is_last_trans = 1
+
         return _pipeline
 
 
